/////////////////////////////////////////////////////////////////////////
// canonical_wps.go
/////////////////////////////////////////////////////////////////////////
// Author: Alexander M Kasprzyk
// Date: 1 April 2013
/////////////////////////////////////////////////////////////////////////
/*
Classifies the canonical weighted projective spaces in low dimensions.
*/

package main

import (
	"bufio"
	"flag"
	"fmt"
	"io"
	"os"
	"runtime"
	"strings"
)

// The command-line arguments structure
type cmdArgs struct {
	n     int    // The dimension to search
	min_h int    // The minimum sum of weights to search from
	max_h int    // The maximum sum of weights to search up to
	out   string // The output file (if any)
}

// A block of memory used to reduce memory assignment
type scratchBuf struct {
	b []int // The chunk of memory
	n int   // The block size
	l int   // The number of remaining blocks
}

// The data describing a job
type jobData struct {
	h     int // The sum of weights
	final int // The final weight
}

// The destination writer
var out_w io.Writer

// The size of the results channel buffer
const CHAN_BUF_SIZE = 16 * 1024

// The size of the scratch buffer
const SCRATCH_SIZE = 64 * 1024

// The size of the output writer buffer
const WRITER_BUF_SIZE = 16 * 1024

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// new_scratch returns a new scratch buffer using block size n + 1.
func new_scratch(n int) *scratchBuf {
	scratch := new(scratchBuf)
	scratch.n = n
	scratch.l = SCRATCH_SIZE
	scratch.b = make([]int, (n+1)*SCRATCH_SIZE)
	return scratch
}

// extract_from_scratch returns a block of memory of length n + 1 from the
// scratch buffer.
func extract_from_scratch(scratch *scratchBuf) []int {
	if scratch.l == 0 {
		scratch.l = SCRATCH_SIZE
		scratch.b = make([]int, (scratch.n+1)*SCRATCH_SIZE)
	}
	a := scratch.b[:scratch.n+1]
	scratch.b = scratch.b[scratch.n+1:]
	scratch.l--
	return a
}

// gcd returns the GCD of the two integers a and b.
func gcd(a int, b int) int {
	if a < 0 {
		return gcd(-a, b)
	}
	if a == 0 {
		return b
	}
	return gcd(b%a, a)
}

// ceiling returns the smallest integer bigger than or equal to a / b.
func ceiling(a int, b int) int {
	c := (b - (a % b)) % b
	return (a + c) / b
}

// floor returns the largest integer smaller than or equal to a / b.
func floor(a int, b int) int {
	c := a % b
	return (a - c) / b
}

// areCoprimeExcluding returns true iff the elements of the given slice are
// coprime, excluding the j-th element.
func areCoprimeExcluding(a []int, j int) bool {
	var b int
	for i, c := range a {
		if i != j {
			if b = gcd(b, c); b == 1 {
				return true
			}
		}
	}
	return false
}

// areWellFormed returns true iff the given weights are well-formed.
func areWellFormed(n int, w []int) bool {
	for i := 0; i <= n; i++ {
		if !areCoprimeExcluding(w, i) {
			return false
		}
	}
	return true
}

// output_weights outputs the weights to the output writer.
func output_weights(done chan bool, out_c chan []int) {
	// Select the output writer
	writer := out_w
	if writer == nil {
		writer = os.Stdout
	}
	// Wrap the destination writer in a buffered writer
	b := bufio.NewWriterSize(writer, WRITER_BUF_SIZE)
	// Start working through the weights
	for w := range out_c {
		// Have we been asked to exit?
		if w == nil {
			break
		}
		// Output the weights
		fmt.Fprintln(b, w)
	}
	// Flush the buffer and indicate that we're finished
	b.Flush()
	done <- true
}

// isCanonical returns true iff the given weights are canonical.
func isCanonical(n int, h int, w []int) bool {
	// Test whether the weights are canonical
	for k := 2; k <= h-2; k++ {
		var s int
		var zeros int
		for _, c := range w {
			if v := (c * k) % h; v == 0 {
				zeros++
			} else {
				s += v
			}
		}
		s /= h
		if (s < 2 || s > n-1) && zeros == 0 {
			return false
		}
	}
	// If we're here then they are
	return true
}

// search_weights_recurse is the recursive heart of the algorithm. Searches
// through all cadidate weights, feeding them to the given channel as they
// are found.
func search_weights_recurse(scratch *scratchBuf, out_c chan []int, w []int,
	h int, k int, dh int) {
	// Check that the starting value is sensible
	if (k+1)*w[k] < dh {
		w[k] = ceiling(dh, k+1)
	}
	dh -= w[k]
	// How we proceed depends on which element k in w we're working with
	n := len(w) - 1
	if k == 1 {
		// The penultimate k = 1 case
		for dh >= 1 && w[1] <= w[2] {
			// We handle the k = 0 case here too
			w[0] = dh
			if areWellFormed(n, w) && isCanonical(n, h, w) {
				// The weights are canonical, so copy to the channel
				new_w := extract_from_scratch(scratch)
				copy(new_w, w)
				out_c <- new_w
			}
			dh--
			w[1]++
		}
	} else {
		// The general case
		max := floor(h, n-k+2)
		for dh >= k && w[k] <= max && w[k] <= w[k+1] {
			search_weights_recurse(scratch, out_c, w, h, k-1, dh)
			dh--
			w[k]++
		}
	}
	// Reset for the next round
	w[k] = 1
}

// search_weights is a wrapper around search_weights_recurse to recursively
// search for all candidate weights of dimension n which sum to h.
func search_weights(done chan bool, job_c chan jobData, out_c chan []int,
	n int) {
	// Allocate the memory for the weights and a scratch buffer
	w := make([]int, n+1)
	scratch := new_scratch(n)
	// Start working through the jobs
	for job := range job_c {
		// Have we been asked to exit?
		if job.h == 0 {
			break
		}
		// Create the initial weights
		for i := 0; i < n; i++ {
			w[i] = 1
		}
		w[n] = job.final
		// Start searching
		search_weights_recurse(scratch, out_c, w, job.h, n-1, job.h-job.final)
	}
	// Indicate that we're finished
	done <- true
}

// search_magager is the main function, coordinating searching for weights.
func search_manager(n int, min_h int, max_h int) {
	// We want to limit the number of searchers
	num_searchers := runtime.NumCPU() + 1
	// Create the local communication channels
	job_c := make(chan jobData, 2*num_searchers)
	job_done := make(chan bool, 1)
	out_c := make(chan []int, CHAN_BUF_SIZE)
	out_done := make(chan bool, 1)
	// Start the output go routine
	go output_weights(out_done, out_c)
	// Launch the searcher go routines
	for i := 0; i < num_searchers; i++ {
		go search_weights(job_done, job_c, out_c, n)
	}
	// Start feeding the searchers
	for h := min_h; h <= max_h; h++ {
		// We subdivide each height
		min := floor(h, n+1)
		if h < (n+1)*min {
			min++
		}
		for i := floor(h, 2); i >= min; i-- {
			job_c <- jobData{h, i}
		}
	}
	// Ask the searchers to exit
	for i := 0; i < num_searchers; i++ {
		job_c <- jobData{0, 0}
	}
	// Wait for the searchers to finish
	finished := 0
	for _ = range job_done {
		finished++
		if finished == num_searchers {
			break
		}
	}
	// Ask the output manager to exit and wait for it to finish
	out_c <- nil
	<-out_done
}

// fatalf outputs the formatted message to Stderr and exits.
func fatalf(format string, a ...interface{}) {
	s := fmt.Sprintf(format, a...)
	fmt.Fprintf(os.Stderr, "%s: %s\n", os.Args[0], strings.TrimSpace(s))
	os.Exit(1)
}

// parse_args parses the command-line arguments.
func parse_args() (args cmdArgs) {
	// Bind the options
	flag.IntVar(&args.n, "n", 3, "The dimension to search in.")
	flag.IntVar(&args.min_h, "hmin", 1, "The minimum height to search from (inclusive).")
	flag.IntVar(&args.max_h, "hmax", 100, "The maximum height to search to (inclusive).")
	flag.StringVar(&args.out, "out", "", "Output weights to the given file.")
	// Parse the arguments
	flag.Parse()
	// Validate the arguments
	if args.n <= 2 {
		fatalf("The dimension must be at least 3.")
	}
	if args.min_h < args.n+1 {
		args.min_h = args.n + 1
	}
	if args.max_h < args.n+1 {
		args.max_h = args.n + 1
	}
	return
}

/////////////////////////////////////////////////////////////////////////
// main
/////////////////////////////////////////////////////////////////////////

// main is the program entry/exit point.
func main() {
	// Use all the machine's cores
	runtime.GOMAXPROCS(runtime.NumCPU())
	// Parse the arguments
	args := parse_args()
	// Create the output file
	if args.out != "" {
		var err error
		if out_w, err = os.Create(args.out); err != nil {
			fatalf("Unable to create output file [%s].", err)
		}
	}
	// Start the search
	search_manager(args.n, args.min_h, args.max_h)
}
